# -*- coding: utf-8 -*-
import getpass
import json
import netrc
import os
import re
import shutil
import stat
import subprocess
import sys
import tarfile
import time
import urlparse
import webbrowser
from StringIO import StringIO

import git
import requests

from .git_sync import GitSyncHandler
from .sync_helpers import extra_git_kwargs, git_pull_develop_bundle, git_update_gitignore
from .utils import (
    bundle_app, bundle_boilerplate, cli_confirm, filter_sass_files, filter_static_files,
    filter_template_files, load_app_config, load_boilerplate_config, resource_path,
    validate_app_config, validate_boilerplate_config
)

CACERT_PEM_PATH = resource_path('cacert.pem')
DATABASE_URL = 'postgres://{user}:{password}@{host}:{port}/{name}'
OPENSSL_DOCKER_CONFLICT_INSTRUCTIONS = """
\n\033[91mYour version of OpenSSL is not compatible with boot2docker and/or docker-python.
Please run the following commands to install a compatible version and try again:
$ brew unlink openssl
$ brew install https://raw.githubusercontent.com/Homebrew/homebrew/62fc2a1a65e83ba9dbb30b2e0a2b7355831c714b/Library/Formula/openssl.rb
$ brew switch openssl 1.0.1j_1
$ brew link openssl --force\033[0m
"""

class WritableNetRC(netrc.netrc):
    def __init__(self, *args, **kwargs):
        netrc_path = self.get_netrc_path()
        if not os.path.exists(netrc_path):
            open(netrc_path, 'a').close()
            os.chmod(netrc_path, 0600)
        # XXX netrc uses os.environ['HOME'] which isn't defined on Windows
        os.environ['HOME'] = os.path.expanduser('~')
        netrc.netrc.__init__(self, *args, **kwargs)

    def get_netrc_path(self):
        home = os.path.expanduser('~')
        return os.path.join(home, '.netrc')

    def add(self, host, login, account, password):
        self.hosts[host] = (login, account, password)

    def remove(self, host):
        if host in self.hosts:
            del self.hosts[host]

    def write(self, path=None):
        if path is None:
            path = self.get_netrc_path()
        with open(path, 'w') as fobj:
            for machine, data in self.hosts.items():
                login, account, password = data
                fobj.write('machine %s\n' % machine)
                if login:
                    fobj.write('\tlogin %s\n' % login)
                if account:
                    fobj.write('\taccount %s\n' % account)
                if password:
                    fobj.write('\tpassword %s\n' % password)


class SingleHostSession(requests.Session):
    def __init__(self, host, **kwargs):
        super(SingleHostSession, self).__init__()
        self.host = host.rstrip('/')
        for key, value in kwargs.items():
            setattr(self, key, value)

    def request(self, method, url, *args, **kwargs):
        url = self.host + url
        # Use local copy of 'cacert.pem' for easier packaging
        kwargs['verify'] = CACERT_PEM_PATH
        return super(SingleHostSession, self).request(method, url, *args, **kwargs)


class Client(object):
    APP_FILENAME_JSON = 'app.json'
    APP_FILENAME_YAML = 'app.yaml'
    ADDON_FILENAME_JSON = 'addon.json'
    ADDON_FILENAME_YAML = 'addon.yaml'
    BOILERPLATE_FILENAME_JSON = 'boilerplate.json'
    BOILERPLATE_FILENAME_YAML = 'boilerplate.yaml'
    ALDRYN_CONFIG_FILENAME = 'aldryn_config.py'
    ALDRYN_DOT_FILENAME = '.aldryn'
    ALDRYN_HOST_DEFAULT = 'https://control.aldryn.com'
    ALDRYN_HOST_KEY = 'ALDRYN_HOST'
    ALDRYN_SYNC_LOCK_FILENAME = '.aldryn-sync-lock'
    DATA_FILENAME = 'data.yaml'
    SETUP_FILENAME = 'setup.py'
    ACCESS_TOKEN_URL_PATH = '/account/desktop-app/access-token/'

    # messages
    DIRECTORY_ALREADY_SYNCING_MESSAGE = 'Directory already syncing.'
    NETWORK_ERROR_MESSAGE = 'Network error.\nPlease check your connection and try again later.'
    PROTECTED_FILE_CHANGE_MESSAGE = 'You are overriding file "%s".\nThis file is protected by the boilerplate.'
    SYNC_NETWORK_ERROR_MESSAGE = "Couldn't sync changes.\nPlease check your connection and try again later."

    ALL_CONFIG_FILES = [
        APP_FILENAME_JSON, APP_FILENAME_YAML,
        BOILERPLATE_FILENAME_JSON, BOILERPLATE_FILENAME_YAML,
        ALDRYN_CONFIG_FILENAME, SETUP_FILENAME, DATA_FILENAME]

    docker_client = None

    @classmethod
    def get_host_url(cls):
        return os.environ.get(cls.ALDRYN_HOST_KEY, cls.ALDRYN_HOST_DEFAULT)

    @classmethod
    def get_access_token_url(cls):
        return '%s/%s' % (
            cls.get_host_url().rstrip('/'), cls.ACCESS_TOKEN_URL_PATH.lstrip('/'))

    def __init__(
            self, host, interactive=True, database_name=None,
            database_host='127.0.0.1', database_port=5432,
            database_user='postgres', database_password=None):
        self.host = urlparse.urlparse(host)[1]
        self.interactive = interactive
        self.netrc = WritableNetRC()
        self.database_name = database_name
        self.database_host = database_host
        self.database_port = database_port
        self.database_user = database_user
        self.database_password = database_password

        auth_data = self.get_auth_data()
        if auth_data:
            headers = {
                'Authorization': 'Basic %s' % auth_data[2]
            }
        else:
            headers = {}
        self.session = SingleHostSession(
            host, headers=headers, trust_env=False)

        self._sync_handlers_cache = {}
        # print 'Connecting to {}'.format(self.host)

    def get_auth_data(self):
        return self.netrc.hosts.get(self.host)

    def is_logged_in(self):
        auth_data = self.get_auth_data()
        if auth_data:
            email, account, token = auth_data
            try:
                response = self.session.post(
                    '/api/v1/verify-token/', data={'token': token})
            except (requests.exceptions.ConnectionError,
                    requests.exceptions.Timeout):
                return False
            else:
                if response.ok:
                    return True
                else:
                    return False
        else:
            return False

    def get_login(self):
        if self.is_logged_in():
            auth_data = self.get_auth_data()
            return auth_data[0]

    def logout(self, force=False):
        while self.interactive and not force:
            answer = raw_input('Are you sure you want to continue? [yN]')
            if answer.lower() == 'n' or not answer:
                sys.stdout.write('Aborted')
                return True
            elif answer.lower() == 'y':
                break
            else:
                sys.stdout.write('Invalid answer, please type either y or n')
        self.netrc.remove(self.host)
        self.netrc.write()

    def login(self, email=None, password=None):
        if email is None:
            email = raw_input('E-Mail: ')
        if password is None:
            password = getpass.getpass('Password: ')
        try:
            response = self.session.post(
                '/api/v1/login/', data={'email': email, 'password': password})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.ok:
            token = response.content
            self.session.headers = {
                'Authorization': 'Basic %s' % token
            }
            self.netrc.add(self.host, email, None, token)
            self.netrc.write()
            msg = "Logged in as %s" % email
            return (True, msg)
        elif response.status_code == requests.codes.forbidden:
            if response.content:
                msg = response.content
            else:
                msg = "Could not log in, invalid email or password"
            return (False, msg)
        else:
            msgs = []
            if response.content and response.status_code < 500:
                msgs.append(response.content)
            msgs.append("There was a problem logging in, please try again later.")
            return (False, '\n'.join(msgs))

    def login_with_token(self, token=None):
        if token is None:
            sys.stdout.write('To get your access token visit: {}'.format(
                self.get_access_token_url()))
            token = raw_input('Access token: ')
        try:
            response = self.session.post(
                '/api/v1/login-with-token/', data={'token': token})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.ok:
            user_data = response.json()
            self.session.headers = {
                'Authorization': 'Basic %s' % token
            }
            email = user_data['email']
            self.netrc.add(self.host, email, None, token)
            self.netrc.write()
            msg = "Logged in as %s" % email
            return (True, msg)
        elif response.status_code == requests.codes.forbidden:
            if response.content:
                msg = response.content
            else:
                msg = "Could not log in, invalid email or password"
            return (False, msg)
        else:
            msgs = []
            if response.content and response.status_code < 500:
                msgs.append(response.content)
            msgs.append("There was a problem logging in, please try again later.")
            return (False, '\n'.join(msgs))

    def upload_boilerplate(self, path='.'):
        is_loaded, result = load_boilerplate_config(path)
        if is_loaded:
            config = result
        else:
            return (False, result)
        is_valid, error_msg = validate_boilerplate_config(config, path)
        if not is_valid:
            return (False, error_msg)
        tarball = bundle_boilerplate(config, path, templates=filter_template_files,
                                     static=filter_static_files, private=filter_sass_files)
        try:
            response = self.session.post('/api/v1/boilerplates/', files={'boilerplate': tarball})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        msg = '\t'.join([str(response.status_code), response.content])
        return (True, msg)

    def validate_boilerplate(self, path='.'):
        is_loaded, result = load_boilerplate_config(path)
        if is_loaded:
            config, extra_file_paths = result
            return validate_boilerplate_config(config, path)
        else:
            return (False, result)

    def upload_app(self, path='.'):
        is_loaded, result = load_app_config(path)
        if is_loaded:
            config = result
        else:
            return (False, result)
        is_valid, msg = validate_app_config(config, path)
        if not is_valid:
            return (False, msg)
        setup_filename = os.path.join(path, Client.SETUP_FILENAME)
        if not os.path.exists(setup_filename):
            msg = "File '%s' not found." % Client.SETUP_FILENAME
            return (False, msg)
        aldryn_config_filename = os.path.join(
            path, Client.ALDRYN_CONFIG_FILENAME)
        msgs = []
        if os.path.exists(aldryn_config_filename):
            with open(aldryn_config_filename) as fobj:
                script = fobj.read()
        else:
            script = ''
            msgs.append("Warning: File '%s' not found." % Client.ALDRYN_CONFIG_FILENAME)
            msgs.append("Your app will not have any configurable settings.")
        tarball = bundle_app(config, script, path)
        try:
            response = self.session.post('/api/v1/apps/', files={'app': tarball})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        msgs.append('\t'.join([str(response.status_code), response.content]))
        return (True, '\n'.join(msgs))

    def validate_app(self, path='.'):
        setup_filename = os.path.join(path, Client.SETUP_FILENAME)
        if not os.path.exists(setup_filename):
            msg = "File '%s' not found." % Client.SETUP_FILENAME
            return (False, msg)
        is_loaded, result = load_app_config(path)
        if is_loaded:
            config = result
            return validate_app_config(config, path)
        else:
            return (False, result)

    def _acquire_sync_lock(self, sync_dir):
        lock_filename = os.path.join(
            sync_dir, Client.ALDRYN_SYNC_LOCK_FILENAME)
        try:
            fd = os.open(lock_filename, os.O_CREAT | os.O_EXCL)
            os.close(fd)
        except OSError:
            return False
        else:
            return True

    def _remove_sync_lock(self, sync_dir):
        lock_filename = os.path.join(
            sync_dir, Client.ALDRYN_SYNC_LOCK_FILENAME)
        if os.path.exists(lock_filename):
            os.remove(lock_filename)

    def sync(self, network_error_callback, sync_error_callback,
             protected_file_change_callback, sitename=None, path='.', force=False,
             sync_indicator_callback=None, stop_sync_callback=None):
        aldryn_dot_filename = os.path.join(path, Client.ALDRYN_DOT_FILENAME)
        if not sitename:
            if os.path.exists(aldryn_dot_filename):
                with open(aldryn_dot_filename, 'r') as fobj:
                    sitename = fobj.read().strip()
            if not sitename:
                msg = "Please specify a sitename using --sitename."
                return (False, msg)
        if '.' in sitename:
            sitename = sitename.split('.')[0]
        git_sync_params = {}
        if not os.path.exists(path):
            os.mkdir(path)
        if os.listdir(path):  # dir isn't empty
            git_dir = os.path.join(path, '.git')
            if os.path.exists(aldryn_dot_filename) and os.path.exists(git_dir):
                with open(aldryn_dot_filename, 'r') as fobj:
                    dir_sitename = fobj.read().strip()
                if sitename != dir_sitename:
                    msg = 'This directory is already being synced with the website "%s"' % dir_sitename
                    return (False, msg)
                repo = git.Repo(path)
                last_synced_commit = repo.git.execute(
                    ['git', 'rev-parse', 'develop_bundle/develop'],
                    **extra_git_kwargs)
                git_sync_params['last_synced_commit'] = last_synced_commit
            else:
                msg = ('The folder you selected is not empty. '
                       'Please choose an empty folder to sync your files.')
                return (False, msg)
        else:
            g = git.Git(path)
            g.execute(['git', 'init'], **extra_git_kwargs)
            repo = git.Repo(path)
            cfg = repo.config_writer()
            cfg.set_value('user', 'name', 'Git Sync by')
            cfg.set_value('user', 'email', self.get_login())
            cfg.write()
            cfg._lock._release_lock()
            del cfg

        if not self._acquire_sync_lock(path):
            if self.interactive:
                if not cli_confirm(
                        'Are you sure you want to start syncing anyway?',
                        message='It seems that you are already syncing this directory.',
                        default=False):
                    return (False, 'Aborted')
            else:
                if force:
                    pass
                else:
                    return (False, Client.DIRECTORY_ALREADY_SYNCING_MESSAGE)

        try:
            response = self.session.get(
                '/api/v1/git-sync/%s/' % sitename, params=git_sync_params, stream=True,
                headers={'accept': 'application/octet'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            self._remove_sync_lock(path)
            return (False, Client.NETWORK_ERROR_MESSAGE)

        def report_unexpected_http_response(response):
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            self._remove_sync_lock(path)
            return (False, '\n'.join(msgs))

        upstream_modified = True
        if response.status_code == 304:
            upstream_modified = False
        elif response.status_code != 200:
            return report_unexpected_http_response(response)

        if upstream_modified:
            try:
                git_pull_develop_bundle(response, repo, path)
            except git.exc.GitCommandError as e:
                # there were changes both in the upstream and uncommited local ones
                # next auto-merge will resolve the conflict (preferring our commits)
                # and then applying the upstream changes
                print e
        git_update_gitignore(repo, ['.*', '!.gitignore', '/db_dumps/'])

        try:
            response = self.session.get(
                '/api/v1/sync/%s/unchanged-protected-files' % sitename,
                headers={'accept': 'application/json'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            self._remove_sync_lock(path)
            return (False, Client.NETWORK_ERROR_MESSAGE)

        if response.status_code != 200:
            return report_unexpected_http_response(response)

        protected_files = response.json()
        # making the protected files read-only
        for rel_protected_path in protected_files:
            protected_path = os.path.join(path, rel_protected_path)
            mode = os.stat(protected_path)[stat.ST_MODE]
            read_only_mode = mode & ~stat.S_IWUSR & ~stat.S_IWGRP & ~stat.S_IWOTH
            os.chmod(protected_path, read_only_mode)

        # successfully initialized the sync, saving the site's name
        with open(aldryn_dot_filename, 'w') as fobj:
            fobj.write(sitename)

        upstream_remote = repo.remotes.develop_bundle
        last_synced_commit = upstream_remote.refs.develop.commit.hexsha

        sync_handler = GitSyncHandler(
            self, sitename, repo, last_synced_commit,
            network_error_callback, sync_error_callback, stop_sync_callback,
            protected_files, protected_file_change_callback,
            relpath=path, sync_indicator_callback=sync_indicator_callback)
        sync_handler.start()
        self._sync_handlers_cache[sitename] = sync_handler
        if self.interactive:
            print "Done, now watching for changes. You can stop the sync by hitting Ctrl-c in this shell"

            try:
                while self._is_syncing(sitename):
                    time.sleep(1)
            except KeyboardInterrupt:
                sync_handler.stop()
            return (True, 'Stopped syncing')
        else:
            return (True, sync_handler)

    def _is_syncing(self, sitename):
        return sitename in self._sync_handlers_cache

    def create_docker_workspace(self, sitename, path=None):
        if path is None:
            path = './%s' % sitename
        path = os.path.abspath(path)
        success, msg = self.workspace_download_site(sitename, path)
        if not success:
            return success, msg
        self.init_docker(sitename, path)
        return True, ''

    def create_workspace(self, sitename, path=None):
        if path is None:
            path = './%s' % sitename
        path = os.path.abspath(path)
        success, msg = self.workspace_download_site(sitename, path=path)
        if not success:
            return success, msg
        success, msg = self.workspace_init_virtualenv(sitename, path)
        if not success:
            return success, msg
        success, msg = self.load_db(sitename, path)
        if not success:
            return success, msg
        return True, sitename

    previous_status = ''

    def print_status(self, status):
        self.previous_status = status
        sys.stdout.write('\033[37m%s' % status)
        sys.stdout.flush()

    def print_done(self):
        sys.stdout.write('\r')
        sys.stdout.write('{0: <25}\033[32m done\n'.format(self.previous_status))
        sys.stdout.write('\033[0m')  # reset color
        sys.stdout.flush()

    def init_docker(self, sitename, path=None):
        from docker.client import Client as DockerClient
        from docker.utils import create_host_config, kwargs_from_env

        self.print_status('Setting up docker')
        site_path = os.path.join(path, '.site')
        os.environ.setdefault('DOCKER_CERT_PATH', CACERT_PEM_PATH)
        docker_kwargs = kwargs_from_env()
        docker_kwargs['tls'].assert_hostname = False
        try:
            self.docker_client = DockerClient(**docker_kwargs)
        except requests.exceptions.SSLError, e:
            sys.stdout.write(OPENSSL_DOCKER_CONFLICT_INSTRUCTIONS)
            raise e

        db_container_name = '{}_db'.format(sitename)
        web_container_name = '{}_web'.format(sitename)

        # Cleanup
        try:
            self.docker_client.stop(db_container_name)
        except:
            pass

        try:
            self.docker_client.stop(web_container_name)
        except:
            pass

        try:
            self.docker_client.remove_container(db_container_name)
        except:
            pass

        try:
            self.docker_client.remove_container(web_container_name)
        except:
            pass

        # Extract 2nd tar.gz which contains dbdump and static files
        with open(os.path.join(site_path, 'backup_archive.tar.gz')) as f:
            with tarfile.open(mode='r:gz', fileobj=f) as tar:
                tar.extract('database.dump', site_path)

        self.print_done()
        self.print_status('Building containers')
        # Setup Containers
        try:
            db_container = self.docker_client.create_container(
                image='postgres:latest',
                volumes='/tmp/data',
                host_config=create_host_config(binds={
                    site_path: {
                        'bind': '/app',
                        'ro': True
                    },
                }),
                name=db_container_name,
            )
        except requests.exceptions.SSLError, e:
            sys.stdout.write(OPENSSL_DOCKER_CONFLICT_INSTRUCTIONS)
            raise e

        # build web container
        self.docker_client.build(
            path=site_path,
            tag='myweb',
            rm=True,
        )

        web_container = self.docker_client.create_container(
            name=web_container_name,
            image='myweb:latest',
            volumes='/tmp/data',
            ports=[80],
            command='/app/launcher.sh',
            host_config=create_host_config(
                binds={
                    site_path: {
                        'bind': '/app',
                        'ro': False
                    },
                },
                port_bindings={80: 8000}
            )
        )
        self.print_done()
        self.print_status('Loading database')
        self.docker_client.start(db_container)

        self.docker_client.exec_create(
            container=db_container,
            cmd='pg_restore -d postgres /app/database.dump',
        )

        self.print_done()
        self.print_status('Starting website')
        self.docker_client.start(
            web_container,
            links={db_container_name: 'db'},
            port_bindings={80: 8000},
        )

        time.sleep(5)

        self.print_done()
        sys.stdout.write("""
\033[32mFinished setting up your local environment,
you may now access your website at:
http://192.168.59.103:8000\033[0m
        """)
        webbrowser.open('http://192.168.59.103:8000')

    def workspace_download_site(self, sitename, path):
        self.print_status('Downloading data')
        if not os.path.exists(path):
            os.mkdir(path)
        site_path = os.path.join(path, '.site')
        try:
            response = self.session.get(
                '/api/v1/workspace/%s/download/' % sitename, stream=True,
                headers={'accept': 'application/x-tar-gz'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code != 200:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))
        else:
            self.print_done()
            # unpack the downloaded tar.gz containing the whole project
            if os.path.exists(site_path):
                # print "deleting old .site"
                shutil.rmtree(site_path)
                if os.path.exists(site_path):
                    return (False, "Failed to delete old .site directory")

            self.print_status('Extracting files')
            tar = tarfile.open(mode='r:gz', fileobj=StringIO(response.content))
            tar.extractall(path=site_path)
            tar.close()
            self.print_done()
            return True, sitename

    def workspace_init_virtualenv(self, sitename, path):
        self.print_status('Setting up virtualenv')
        virtualenv_path = os.path.join(path, '.virtualenv')
        requirements_path = os.path.join(path, '.site/requirements.txt')
        pip_path = os.path.join(path, '.virtualenv/bin/pip')
        dev_path = os.path.join(path, 'dev')

        if not os.path.exists(dev_path):
            os.mkdir(dev_path)

        # fix requirements
        with open(requirements_path, 'r') as f:
            lines = f.readlines()

        excluded = ['mercurial', 'bzr', 'gyp', 'pygobject', 'pygpgme', 'cffi']

        with open(requirements_path, 'w') as f:
            for line in lines:
                if not any(re.match(pat, line) for pat in excluded):
                    f.write(line)

        self.print_done()
        self.print_status('Installing requirements')
        subprocess.call(['virtualenv', virtualenv_path], stdout=open(os.devnull, 'wb'))
        subprocess.call([pip_path, 'install', 'pyOpenSSL==0.15.1', 'ndg-httpsclient==0.3.3', 'pyasn1==0.1.7', 'cryptography==0.8.2'], stdout=open(os.devnull, 'wb'))
        subprocess.call([pip_path, 'install', '--allow-all-external', '--allow-unverified', 'lazr.restfulclient', '-r', requirements_path], stdout=open(os.devnull, 'wb'))
        with open(os.path.join(virtualenv_path, 'lib/python2.7/site-packages/aldrynsite.pth'), 'w+') as fobj:
            fobj.write(os.path.abspath('.site/'))
        with open(os.path.join(virtualenv_path, 'lib/python2.7/site-packages/aldrynsite_dev.pth'), 'w+') as fobj:
            fobj.write(os.path.abspath(dev_path))
        self.print_done()
        return True, sitename

    def load_db(self, sitename, path):
        self.print_status('Fetching database')
        site_path = os.path.join(path, '.site')
        tmp_path = os.path.join(path, '.tmp')
        if not os.path.exists(tmp_path):
            os.mkdir(tmp_path)
        try:
            response = self.session.get(
                '/api/v1/workspace/%s/download/db/' % sitename, stream=True,
                headers={'accept': 'application/x-tar-gz'})
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code != 200:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))

        self.print_done()

        # unpack the downloaded tar.gz containing the database dump
        self.print_status('Extracting files')
        from StringIO import StringIO
        tar = tarfile.open(mode='r:gz', fileobj=StringIO(response.content))
        tar.extractall(path=tmp_path)
        tar.close()
        self.print_done()

        # Set up .env
        self.print_status('Setting up .env')
        database_name = self.database_name or 'aldryn_{}'.format(sitename)
        database_url = DATABASE_URL.format(
            user=self.database_user, password=self.database_password,
            host=self.database_host, port=self.database_port,
            name=database_name
        )

        with open(os.path.join(site_path, '.env'), 'w') as f:
            f.write('DATABASE_URL={}'.format(database_url))
        self.print_done()

        self.print_status('Loading database')
        # delete old database
        subprocess.call([
            'psql',
            '-h', self.database_host,
            '-p', self.database_port,
            '-U', self.database_user,
            '-c', 'DROP DATABASE "{}"'.format(database_name)
        ], stdout=open(os.devnull, 'wb'))

        # create fresh new
        subprocess.call([
            'psql',
            '-h', self.database_host,
            '-p', self.database_port,
            '-U', self.database_user,
            '-c', 'CREATE DATABASE "{}"'.format(database_name)
        ], stdout=open(os.devnull, 'wb'))

        # load data
        subprocess.call([
            'pg_restore',
            '-h', self.database_host,
            '-p', self.database_port,
            '-U', self.database_user,
            '-d', database_name,
            '--no-owner',
            os.path.join(tmp_path, 'database.dump'),
        ],  stdout=open(os.devnull, 'wb'))
        self.print_done()

        return True, sitename

    def sites(self):
        try:
            response = self.session.get('/api/v1/sites/', stream=True)
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code == 200:
            sites = json.loads(response.content)
            if self.interactive:
                data = json.dumps(sites, sort_keys=True, indent=4, separators=(',', ': '))
            else:
                data = sites
            return (True, data)
        elif response.status_code == 403:
            self.logout(force=True)
            msg = 'Session expired. Please log in again.'
            return (False, msg)
        else:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))

    def newest_version(self):
        try:
            response = self.session.get('/api/v1/newest-client-app-version/')
        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout):
            return (False, Client.NETWORK_ERROR_MESSAGE)
        if response.status_code == 404:
            return (True, None)
        elif response.status_code != 200:
            msgs = []
            msgs.append("Unexpected HTTP Response %s" % response.status_code)
            if response.status_code < 500:
                msgs.append(response.content)
            return (False, '\n'.join(msgs))
        else:
            version = json.loads(response.content)
            return (True, version)



# -*- coding: utf-8 -*-
# Configuring kivy
# window's sizes and position
WIDTH_LOGIN = 400
HEIGHT_LOGIN = 600
WIDTH_SYNC = 1200
HEIGHT_SYNC = 700

import os
from kivy.config import Config
Config.set('kivy', 'desktop', 1)
# disable multitouch emulation with mouse's clicks
Config.set('input', 'mouse', 'mouse,disable_multitouch')
from .utils import get_icon_path
Config.set('kivy', 'window_icon', get_icon_path())
Config.write()

# Kivy's import must happen after the configuration
import kivy
kivy.require('1.7.2')

from distutils.version import LooseVersion
from functools import partial
import anydbm
import platform
import shelve
import traceback
import webbrowser

from kivy.app import App
from kivy.base import stopTouchApp
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.uix.popup import Popup
from kivy.uix.screenmanager import ScreenManager, TransitionBase
import pygame

from . import __version__ as current_version
from .client import Client
from .utils_kivy import (
    ConfirmDialog,
    DirChooserDialog,
    EmptyScreen,
    HOME_DIR,
    InfoDialog,
    LinkButton,
    LoadSitesListThread,
    LoadingDialog,
    LoginScreen,
    LoginThread,
    LoginWithTokenThread,
    SyncDirThread,
    SyncScreen,
    WebsitesManager,
    notify,
)

KNOWN_CONFIG_FILES_FILTERS = ['*.yaml', '*.py']
LAST_DIR_KEY = 'last_dir'
SITES_DATABASE_FILENAME = 'sites_database'
USER_SETTINGS_SECTION = 'user_settings'
WINDOW_TITLE = 'Aldryn App'

# window resizing parameters
RESIZING_DURATION = 0.0
RESIZING_EASING = 7
RESIZING_STEPS = 1

# urls
ACCOUNT_CREATION_URL = 'https://login.django-cms.com/login/'
TROUBLE_SIGNING_IN_URL = 'https://login.django-cms.com/account/reset-password/'
CONTROL_PANEL_URL = 'https://control.django-cms.com/control/'
ADD_NEW_SITE_URL = 'https://control.django-cms.com/control/choose-website-type/'


####################
# App's main class #
####################

class AldrynGUIApp(App):

    title = WINDOW_TITLE
    icon = get_icon_path()

    def __init__(self, *args, **kwargs):
        super(AldrynGUIApp, self).__init__(*args, **kwargs)
        self.client = Client(Client.get_host_url(), interactive=False)
        sites_dir_database_file_path = os.path.join(self.user_data_dir, SITES_DATABASE_FILENAME)

        def open_db():
            self.sites_dir_database = shelve.open(sites_dir_database_file_path, writeback=True)
        try:
            open_db()
        except anydbm.error:  # "db type could not be determined"
            traceback.print_exc()  # log the exception
            os.remove(sites_dir_database_file_path)  # remove corrupted db file
            open_db()  # retry

    def build_config(self, config):
        config.setdefaults(USER_SETTINGS_SECTION, {
            LAST_DIR_KEY: HOME_DIR
        })

    def build(self):
        sm = ScreenManager(transition=TransitionBase(duration=0.1))
        sm.add_widget(LoginScreen(name='login'))
        sm.add_widget(SyncScreen(name='sync'))
        sm.add_widget(EmptyScreen(name='empty'))
        return sm

    def on_key_down(self, win, key, scancode, string, modifiers):
        system = platform.system()
        if system == 'Darwin':
            if 'meta' in modifiers:
                if key == 113:  # cmd+q
                    stopTouchApp()
                    return True
                if key == 104:  # cmd+h
                    pygame.display.iconify()
                    return True
        elif platform == 'Windows':
            if 'alt' in modifiers and key == 285:  # alt+f4
                stopTouchApp()
                return True

    def on_start(self):
        super(AldrynGUIApp, self).on_start()

        Window.bind(on_key_down=self.on_key_down)

        sites_list_view = self._get_sites_list_view()
        self._websites_manager = WebsitesManager(self.sites_dir_database, sites_list_view)

        if self.client.is_logged_in():
            self.set_screen_to_sync()
        else:
            self.set_screen_to_login()

    def on_stop(self):
        self.sites_dir_database.close()
        self.config.write()
        self._websites_manager.stop_all_threads()
        super(AldrynGUIApp, self).on_stop()

    def set_screen_to_sync(self):
        Window.size = WIDTH_SYNC, HEIGHT_SYNC
        self.root.current = 'sync'
        self.load_sites_list()

    def set_screen_to_login(self):
        Window.size = WIDTH_LOGIN, HEIGHT_LOGIN
        self.root.current = 'login'

    ### DIALOGS ###

    def show_info_dialog(self, title, msg, on_open=None, on_close=None):
        info_popup = None

        def dismiss_info_dialog():
            if info_popup and hasattr(info_popup, 'dismiss'):
                info_popup.dismiss()
            if callable(on_close):
                on_close()

        content = InfoDialog(close=dismiss_info_dialog)
        content.message_label.text = msg
        info_popup = Popup(
            title=title, content=content, auto_dismiss=False,
            size_hint=(0.9, None), height=200)
        if callable(on_open):
            info_popup.on_open = on_open
        info_popup.open()
        return info_popup

    def show_confirm_dialog(self, title, msg, on_confirm, on_cancel=None,
                            cancel_btn_text='Cancel',
                            confirm_btn_text='Confirm',
                            on_open=None):
        confirm_popup = None

        def dismiss_confirm_dialog():
            if confirm_popup and hasattr(confirm_popup, 'dismiss'):
                confirm_popup.dismiss()

        def on_confirm_wrapper():
            dismiss_confirm_dialog()
            on_confirm()

        def on_cancel_wrapper():
            dismiss_confirm_dialog()
            if callable(on_cancel):
                on_cancel()

        content = ConfirmDialog(confirm_callback=on_confirm_wrapper,
                                cancel_callback=on_cancel_wrapper,
                                cancel_btn_text=cancel_btn_text,
                                confirm_btn_text=confirm_btn_text)
        content.message_label.text = msg
        confirm_popup = Popup(
            title=title, content=content,
            auto_dismiss=False, size_hint=(0.9, None), height=200)
        if on_open:
            confirm_popup.on_open = on_open
        confirm_popup.open()
        return confirm_popup

    def show_loading_dialog(self, on_open=None):
        content = LoadingDialog()
        loading_popup = Popup(
            title='', auto_dismiss=False, content=content,
            size_hint=(0.9, None), height=200)
        if on_open:
            loading_popup.on_open = on_open
        loading_popup.open()
        return loading_popup

    def show_dir_chooser_dialog(self, on_selection, path=None, on_open=None):
        dir_chooser_popup = None

        def dismiss_dir_chooser_dialog():
            if dir_chooser_popup and hasattr(dir_chooser_popup, 'dismiss'):
                dir_chooser_popup.dismiss()

        def on_selection_wrapper(path, selection, new_dir_name):
            dir_path = path
            if new_dir_name:
                dir_path = os.path.join(path, new_dir_name)
                try:
                    os.mkdir(dir_path)
                except OSError as e:
                    if e.errno == 17:  # File exists
                        msg = 'Directory "%s" already exists' % new_dir_name
                    else:
                        msg = 'Invalid filename'
                    self.show_info_dialog('Error', msg)
                    return
            elif selection:
                if selection[0].startswith('..'):  # disallow selection of the parent directory
                    return
                else:
                    dir_path = selection[0]
            dismiss_dir_chooser_dialog()
            on_selection(dir_path)

        content = DirChooserDialog(
            select=on_selection_wrapper, cancel=dismiss_dir_chooser_dialog)
        file_chooser = content.file_chooser
        file_chooser.path = path or self._get_last_dir()
        file_chooser.bind(path=lambda instance, path: self._set_last_dir(path))
        dir_chooser_popup = Popup(
            title="Choose directory", content=content, size_hint=(0.9, 0.9))
        if on_open:
            dir_chooser_popup.on_open = on_open
        dir_chooser_popup.open()
        return dir_chooser_popup

    ### END OF DIALOGS ###

    def _resize_window(self, target_width, target_height, callback=None):
        current_width, current_height = Window.size
        step_duration = RESIZING_DURATION / RESIZING_STEPS

        def set_size(steps_left, dt):
            steps_left = steps_left - 1 if steps_left > 0 else 0
            ratio = steps_left * 1.0 / RESIZING_STEPS
            ratio **= RESIZING_EASING
            width = int(current_width * ratio + target_width * (1.0 - ratio))
            height = int(current_height * ratio + target_height * (1.0 - ratio))
            Window.size = width, height
            if steps_left > 0:
                Clock.schedule_once(partial(set_size, steps_left), step_duration)
            elif callback:
                callback()

        set_size(RESIZING_STEPS, 0.0)

    def login(self, email, password):
        loading_dialog = self.show_loading_dialog()
        login_callback = partial(self._login_callback, loading_dialog)
        self._login_thread = LoginThread(
            email, password, self.client, login_callback)
        self._login_thread.start()

    def login_with_token(self, token):
        loading_dialog = self.show_loading_dialog()
        login_callback = partial(self._login_callback, loading_dialog)
        self._login_thread = LoginWithTokenThread(
            token, self.client, login_callback)
        self._login_thread.start()

    def _login_callback(self, loading_dialog, status, msg):
        loading_dialog.dismiss()
        if status:
            # Removing the credentials from the inputs of the login's screen
            login_screen = self.root.get_screen('login')
            login_screen.reset_inputs()

            def callback():  # change screen to sync after the animation
                self.root.get_screen('sync').on_enter = self.load_sites_list
                self.root.current = 'sync'
            # start the animation after the empty screen is loaded
            animate_window_resize = partial(self._resize_window, WIDTH_SYNC, HEIGHT_SYNC, callback=callback)
            self.root.get_screen('empty').on_enter = animate_window_resize
            self.root.current = 'empty'
        else:
            notify(WINDOW_TITLE, msg, bring_up=True)
            self.show_info_dialog('Error', msg)

    def logout(self):
        self.client.logout()
        self._websites_manager.clear_websites()

        def callback():  # change screen from empty to sync after the animation
            self.root.current = 'login'
        # start the animation after the empty screen is loaded
        animate_window_resize = partial(self._resize_window, WIDTH_LOGIN, HEIGHT_LOGIN, callback=callback)
        self.root.get_screen('empty').on_enter = animate_window_resize
        self.root.current = 'empty'

    def _get_sites_list_view(self):
        return self.root.get_screen('sync').sites_list_view

    def load_sites_list(self):
        loading_dialog = self.show_loading_dialog()
        load_sites_list_callback = partial(
            self._load_sites_list_callback, loading_dialog)
        self._load_sites_list_thread = LoadSitesListThread(
            self.client, load_sites_list_callback)
        self._load_sites_list_thread.start()

    def _load_sites_list_callback(self, loading_dialog, status, data, version_data=None):
        loading_dialog.dismiss()
        if status:
            new_sites_names = set()
            old_sites_names = set(self._websites_manager.get_domains())
            for site_data in data:
                domain = site_data['domain'].encode('utf-8')
                new_sites_names.add(domain)
                self._websites_manager.add_or_update_website(
                    domain, site_data)
            removed_sites_names = old_sites_names - new_sites_names
            for removed_domain in removed_sites_names:
                self._websites_manager.remove_website(removed_domain)

            if version_data:
                newest_version = version_data['version']
                if LooseVersion(newest_version) > LooseVersion(current_version):
                    system = platform.system()
                    if system == 'Darwin':
                        link = version_data['osx_link']
                    elif system == 'Windows':
                        link = version_data['windows_link']
                    elif system == 'Linux':
                        if platform.architecture().startswith('32'):
                            link = version_data['linux32_link']
                        else:
                            link = version_data['linux64_link']
                    else:
                        link = None
                    if link:
                        sync_screen = self.root.get_screen('sync')
                        if hasattr(sync_screen, 'newest_version_btn'):
                            newest_version_btn = sync_screen.newest_version_btn
                        else:
                            newest_version_btn = LinkButton()
                            sync_screen.newest_version_btn = newest_version_btn
                            sync_screen.top_right_buttons.add_widget(newest_version_btn)
                        newest_version_btn.text = 'Newer version is available (%s => %s)' % (current_version, newest_version)
                        newest_version_btn.bind(on_release=lambda *args: webbrowser.open(link))

        else:
            msg = unicode(data)
            self.show_info_dialog('Error', msg)

    def select_site_dir(self, domain, on_selection=None):

        def on_selection_wrapper(site_dir):
            self._select_site_dir_callback(domain, site_dir)
            if callable(on_selection):
                on_selection()
        site_dir = self._websites_manager.get_site_dir(domain)
        self.show_dir_chooser_dialog(on_selection_wrapper, path=site_dir)

    def _select_site_dir_callback(self, domain, site_dir):
        self._websites_manager.set_site_dir(domain, site_dir)

    ### SYNC ###

    def sync_toggle(self, domain):
        handler = self._websites_manager.get_site_sync_handler(domain)
        if handler:
            self.stop_sync(domain)
        else:
            site_dir = self._websites_manager.get_site_dir(domain)
            if site_dir:
                # XXX since git sync solution we don't discard any local changes
                # on_confirm = partial(self._sync_confirmed, domain, site_dir)
                # title = 'Confirm sync'
                # name = self._websites_manager.get_site_name(domain)
                # msg = 'All local changes to the boilerplate of "%s" will be undone. Continue?' % name
                # self.show_confirm_dialog(title, msg, on_confirm)
                self._sync_confirmed(domain, site_dir)
            else:
                self.select_site_dir(domain, on_selection=lambda *args: self.sync_toggle(domain))

    def stop_sync(self, domain):
        self._websites_manager.stop_site_sync_handler(domain)

    def _sync_confirmed(self, domain, site_dir, force=False):
        loading_dialog = self.show_loading_dialog()
        path = site_dir.encode('utf-8')
        sync_callback = partial(self._sync_callback, loading_dialog)
        stop_sync_callback = partial(self._stop_sync_callback, domain)

        def network_error_callback(message, on_confirm, on_cancel):
            notify(WINDOW_TITLE, message, bring_up=True)

            def on_cancel_wrapper():
                self.stop_sync(domain)
                on_cancel()
            self.show_confirm_dialog(
                'Network error', message,
                on_confirm, on_cancel=on_cancel_wrapper,
                confirm_btn_text='Retry',
                cancel_btn_text='Stop sync (lose unsynced changes!)')

        def sync_error_callback(message, title='Error'):
            notify(WINDOW_TITLE, message, bring_up=True)
            return self.show_info_dialog(title, message)

        def protected_file_change_callback(message):
            notify(WINDOW_TITLE, message)
            self.show_info_dialog('Info', message)

        def sync_indicator_callback(stop=False):
            if stop:
                self._websites_manager.hide_sync_loading_overlay(domain)
            else:
                self._websites_manager.show_sync_loading_overlay(domain)

        sync_dir_thread = SyncDirThread(
            domain, path, force, self.client,
            sync_callback,
            sync_indicator_callback,
            stop_sync_callback,
            network_error_callback,
            sync_error_callback,
            protected_file_change_callback)
        sync_dir_thread.start()

    def _sync_callback(self, loading_dialog, domain, status, msg_or_sync_handler):
        loading_dialog.dismiss()
        if status:  # handler
            self._websites_manager.set_site_sync_handler(domain, msg_or_sync_handler)
        else:  # msg
            if msg_or_sync_handler == Client.DIRECTORY_ALREADY_SYNCING_MESSAGE:
                site_dir = self._websites_manager.get_site_dir(domain)
                if site_dir:
                    on_confirm = partial(
                        self._sync_confirmed, domain, site_dir, force=True)
                    self.show_confirm_dialog(
                        'Directory already syncing',
                        'It seems that you are already syncing this directory',
                        on_confirm,
                        cancel_btn_text='Cancel',
                        confirm_btn_text='Continue anyway')
                else:
                    self.select_site_dir(domain)
            else:
                self.show_info_dialog('Error', msg_or_sync_handler)

    def _stop_sync_callback(self, domain, msg, force=False, logout=False):
        handler = self._websites_manager.get_site_sync_handler(domain)
        if handler:
            if hasattr(handler, '_continue_despite_stop_callback'):
                return
            else:
                handler._continue_despite_stop_callback = True
        notify(WINDOW_TITLE, msg)

        def on_done():
            self.stop_sync(domain)
            if logout:
                self.logout()
        if force:
            title = 'Syncing stopped'
            self.show_info_dialog(title, msg, on_close=on_done)
        else:
            title = 'Stop syncing'
            cancel_btn_text = 'No, continue'
            confirm_btn_text = 'Yes, stop syncing'
            self.show_confirm_dialog(title, msg, on_done,
                                     cancel_btn_text=cancel_btn_text,
                                     confirm_btn_text=confirm_btn_text)

    ### END OF SYNC ###

    def _set_last_dir(self, path):
        self.config.set(USER_SETTINGS_SECTION, LAST_DIR_KEY, path)
        self.config.write()

    def _get_last_dir(self):
        return self.config.getdefault(USER_SETTINGS_SECTION, LAST_DIR_KEY, HOME_DIR)

    def browser_open_account_creation(self):
        webbrowser.open(ACCOUNT_CREATION_URL)

    def browser_open_trouble_signing_in(self):
        webbrowser.open(TROUBLE_SIGNING_IN_URL)

    def browser_open_control_panel(self):
        webbrowser.open(CONTROL_PANEL_URL)

    def browser_open_add_new_site(self):
        webbrowser.open(ADD_NEW_SITE_URL)

    def browser_open_site_dashboard(self, domain):
        webbrowser.open(self._websites_manager.get_site_dashboard_url(domain) or
                        CONTROL_PANEL_URL)

    def browser_open_stage(self, domain):
        webbrowser.open(self._websites_manager.get_site_stage_url(domain) or
                        CONTROL_PANEL_URL)

    def browser_open_get_access_token(self):
        webbrowser.open(self.client.get_access_token_url())

if __name__ == '__main__':
    AldrynGUIApp().run()



"""
This is a setup.py script generated by py2applet

Usage:
    python setup.py py2app
"""

from setuptools import setup

APP = ['main.py']
DATA_FILES = []
OPTIONS = {
    'excludes': ['numpy', 'email'],
    'argv_emulation': False,
    'iconfile': 'resources/appIcon.icns',
    'semi_standalone': 'False',
    'optimize': 2,
    'resources': ['client.py', 'sync.py', 'utils_kivy.py', 'utils.py', 'cacert.pem', 'img', 'management', 'aldryngui.kv'],
    'packages': ['certifi', 'wxwidgets', 'wxpython'],
    'site_packages': 'True',
    'plist': {
        'LSPrefersPPC': False,
        'LSArchitecturePriority': 'x86_64',
    }
}

setup(
    name='AldrynCloud',
    version='0.0.1',
    author='Divio AG',
    author_email='info@divio.ch',
    description='App to control static files of Aldryn Cloud sites',
    app=APP,
    data_files=DATA_FILES,
    options={'py2app': OPTIONS},
    setup_requires=['py2app'],
)



# -*- coding: utf-8 -*-ALDRYN_DUMPDATA_FOLLOW
import os
from django.core.management.base import NoArgsCommand
from aldryn_client.serialize import Loader


class Command(NoArgsCommand):
    def handle(self, infile=None, language=None, **options):
        if infile is None:
            print "You must specify a file to read from as first argument"
            return
        if language is None:
            print "You must specify a language as second argument"
            return
        loader = Loader(language)
        loader.load(infile)



# -*- coding: utf-8 -*-
import imp
import platform
import sys

import docopt

from . import __version__ as version
from .client import Client
from .utils import cli_confirm

try:
    imp.find_module('kivy')
except ImportError:
    GUI = False
else:
    GUI = True

doc_draft = """Aldryn client.

Usage:%(extra_commands)s
    aldryn login [--with-token]
    aldryn boilerplate upload
    aldryn boilerplate validate
    aldryn addon upload
    aldryn addon validate
    aldryn sync [--sitename=<sitename>]
    aldryn workspace create --sitename=<sitename> [--path=<sitename>] [--docker] [--db-name=<database_name>]  [--db-host=<database_host>] [--db-port=<database_port>] [--db-user=<database_user>] [--db-pass=<database_password>]
    aldryn sites
    aldryn newest_version

Options:
    -h --help                        Show this screen.
    --version                        Show version.
    --sitename=<sitename>            Name of your site, eg your-project-name
    --db-name=<database_name>        Database Name
    --db-host=<database_host>        Database Host [default: 127.0.0.1]
    --db-port=<database_port>        Database Port [default: 5432]
    --db-user=<database_user>        Database User [default: postgres]
    --db-pass=<database_password>    Database Password [default: '']
"""

gui_command = """
    aldryn gui"""

if GUI:
    __doc__ = doc_draft % {'extra_commands': gui_command}
else:
    __doc__ = doc_draft % {'extra_commands': ''}


def _network_error_callback(message, on_confirm, on_cancel):
    question = 'Retry syncing the file?'
    if cli_confirm(question, message=message, default=True):
        on_confirm()
    else:
        on_cancel()


def _protected_file_change_callback(message):
    print '!!!'
    print message
    print '!!!'


def _sync_error_callback(message, title=None):
    if title:
        print title
    print message


def main():
    args = docopt.docopt(__doc__, version=version)
    client = Client(
        Client.get_host_url(),
        interactive=True,
        database_name=args['--db-name'],
        database_host=args['--db-host'],
        database_port=args['--db-port'],
        database_user=args['--db-user'],
        database_password=args['--db-pass'],
    )
    retval = True
    msg = None
    if GUI and args['gui']:
        from main import AldrynGUIApp
        AldrynGUIApp().run()
    elif args['login']:
        if args['--with-token']:
            retval, msg = client.login_with_token()
        else:
            retval, msg = client.login()
    elif args['boilerplate']:
        if args['upload']:
            retval, msg = client.upload_boilerplate()
        elif args['validate']:
            retval, msg = client.validate_boilerplate()
    elif args['addon']:
        if args['upload']:
            retval, msg = client.upload_app()
        elif args['validate']:
            retval, msg = client.validate_app()
    elif args['sync']:
        retval, msg = client.sync(
            _network_error_callback,
            _sync_error_callback,
            _protected_file_change_callback,
            sitename=args.get('--sitename', None))
    elif args['sites']:
        retval, msg = client.sites()
    elif args['newest_version']:
        print 'Current version: %s' % version
        retval, version_data = client.newest_version()
        if retval:
            if version_data:
                newest_version = version_data['version']
            else:
                newest_version = None
            if newest_version and newest_version > version:
                system = platform.system()
                if True or system == 'Darwin':
                    link = version_data['osx_link']
                elif system == 'Windows':
                    link = version_data['windows_link']
                elif system == 'Linux':
                    if platform.architecture().startswith('32'):
                        link = version_data['linux32_link']
                    else:
                        link = version_data['linux64_link']
                else:
                    link = None
                if link:
                    msg = 'You can download the newest version (%s) from here:\n%s' % (
                        newest_version, link)
                else:
                    msg = 'Newer version is available (%s).' % newest_version
            else:
                msg = 'You are using the latest version.'
    elif args['workspace']:
        if args['create']:
            if args.get('--docker'):
                try:
                    import docker
                except ImportError:
                    exit(
                        "\033[91mIn order to use the experimental Docker "
                        "support you need to install the package with docker "
                        "dependencies: pip install aldryn-client[docker]"
                        "\033[0m"
                    )
                sys.stdout.write(
                    "\033[91mCaution: Docker support is still experimental!"
                    "\033[0m\n"
                )
                retval, msg = client.create_docker_workspace(
                    sitename=args.get('--sitename', None),
                    path=args.get('--path', None),
                )
            else:
                retval, msg = client.create_workspace(
                    sitename=args.get('--sitename', None),
                    path=args.get('--path', None),
                )
    if msg:
        print '\n{}'.format(msg)
    sys.exit(int(retval))



# -*- coding: utf-8 -*-
import os
import platform
import subprocess
import threading

from kivy.app import App
from kivy.clock import Clock
from kivy.logger import Logger
from kivy.properties import ObjectProperty, StringProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.button import Button
from kivy.uix.filechooser import FileChooserListView
from kivy.uix.image import Image
from kivy.uix.relativelayout import RelativeLayout
from kivy.uix.screenmanager import Screen
from kivy.uix.textinput import TextInput
from kivy.utils import get_color_from_hex
from plyer import notification

from .utils import resource_path, get_icon_path

HOME_DIR = os.path.expanduser('~')
# status to dot color mapping
STATUS_TO_COLOR = {
    0: get_color_from_hex('#4a9cf2'),  # NEW, blue
    1: get_color_from_hex('#98ba0f'),  # DEPLOYED, green
    2: get_color_from_hex('#4a9cf2'),  # DEPLOYING, blue
    3: get_color_from_hex('#626262'),  # OFFLINE, grey
    4: get_color_from_hex('#f0453b'),  # ERROR, red
}


###########
# Dialogs #
###########

class InfoDialog(BoxLayout):
    close = ObjectProperty(None)
    message_label = ObjectProperty(None)


class ConfirmDialog(BoxLayout):
    cancel_callback = ObjectProperty(None)
    confirm_callback = ObjectProperty(None)
    message_label = ObjectProperty(None)
    cancel_btn_text = ObjectProperty(None)
    confirm_btn_text = ObjectProperty(None)


class LoadingDialog(RelativeLayout):
    message_label = ObjectProperty(None)


class DirChooserDialog(BoxLayout):
    cancel = ObjectProperty(None)
    file_chooser = ObjectProperty(None)
    select = ObjectProperty(None)


##################
# Custom widgets #
##################

class TabTextInput(TextInput):
    next = ObjectProperty(None)

    def _keyboard_on_key_down(self, window, keycode, text, modifiers):
        key, key_str = keycode
        if key in (9, 13) and self.next is not None:
            if isinstance(self.next, TextInput):
                self.next.focus = True
                self.next.select_all()
            if isinstance(self.next, Button):
                self.focus = False
                self.next.dispatch('on_press')
                self.next.dispatch('on_release')
        else:
            super(TabTextInput, self)._keyboard_on_key_down(window, keycode, text, modifiers)


class LoadingImage(Image):
    loader_img = resource_path('img/loader.gif')


class PaddedButton(Button):
    pass


class LinkButton(PaddedButton):
    pass


class OpenButton(Button):
    external_img = resource_path('img/external.png')


class CustomFileChooserListView(FileChooserListView):

    def __init__(self, *args, **kwargs):
        super(CustomFileChooserListView, self).__init__(*args, **kwargs)
        self.bind(path=self._normalize_path)

    def _normalize_path(self, *args, **kwargs):
        norm_path = os.path.normpath(self.path)
        if norm_path != self.path:
            self.path = norm_path

    def go_to_home_dir(self):
        self.path = HOME_DIR


class WebsiteView(RelativeLayout):
    status_btn = ObjectProperty(None)
    name_btn = ObjectProperty(None)
    dir_label = ObjectProperty(None)
    change_or_set_dir_btn = ObjectProperty(None)
    open_dir_btn = ObjectProperty(None)
    preview_btn = ObjectProperty(None)
    sync_btn = ObjectProperty(None)
    sync_loading_overlay = ObjectProperty(None)

    domain = StringProperty(None)

    def __init__(self, domain):
        super(WebsiteView, self).__init__()
        self.domain = domain
        self.hide_sync_loading_overlay()

    def set_name(self, name):
        self.name_btn.text = name

    def set_status_color(self, color):
        self.status_btn.color = color

    def _set_dir_btn_text_to_change(self):
        self.change_or_set_dir_btn.text = 'change'

    def _set_dir_btn_text_to_set(self):
        self.change_or_set_dir_btn.text = 'Set sync destination folder'

    def set_site_dir_widgets(self, site_dir):
        if site_dir:
            self.dir_label.text = site_dir
            self._set_dir_btn_text_to_change()
            if not self.open_dir_btn:
                open_dir_btn = OpenButton()
                self.change_or_set_dir_btn.parent.add_widget(open_dir_btn)
                self.open_dir_btn = open_dir_btn
            self.open_dir_btn.on_release = lambda: open_in_file_manager(site_dir)
        else:
            self._set_dir_btn_text_to_set()

    def set_sync_btn_text_to_stop(self):
        if not hasattr(self, '_background_color_orig'):
            self._background_color_orig = self.sync_btn.background_color
        if not hasattr(self, '_background_down_orig'):
            self._background_down_orig = self.sync_btn.background_down
        if not hasattr(self, '_background_normal_orig'):
            self._background_normal_orig = self.sync_btn.background_normal

        self.sync_btn.background_color = (0.33, 0.66, 0.9, 1)
        self.sync_btn.background_down = ''
        self.sync_btn.background_normal = ''
        self.sync_btn.text = 'Stop Sync'

    def set_sync_btn_text_to_sync(self):
        if hasattr(self, '_background_color_orig'):
            self.sync_btn.background_color = self._background_color_orig
        if hasattr(self, '_background_down_orig'):
            self.sync_btn.background_down = self._background_down_orig
        if hasattr(self, '_background_normal_orig'):
            self.sync_btn.background_normal = self._background_normal_orig

        self.sync_btn.text = 'Sync Files'

    def show_sync_loading_overlay(self):
        if self.sync_loading_overlay.parent is None:
            self.sync_btn.add_widget(self.sync_loading_overlay)

    def hide_sync_loading_overlay(self):
        if self.sync_loading_overlay.parent is not None:
            self.sync_btn.remove_widget(self.sync_loading_overlay)


######################
# Asynchronous tasks #
######################

class LoginThread(threading.Thread):

    def __init__(self, email, password, client, callback):
        super(LoginThread, self).__init__()
        self.email = email
        self.password = password
        self.client = client
        self.callback = callback

    def run(self):
        status, msg = self.client.login(email=self.email, password=self.password)
        Clock.schedule_once(lambda dt: self.callback(status, msg), 0)


class LoginWithTokenThread(threading.Thread):

    def __init__(self, token, client, callback):
        super(LoginWithTokenThread, self).__init__()
        self.token = token
        self.client = client
        self.callback = callback

    def run(self):
        status, msg = self.client.login_with_token(token=self.token)
        Clock.schedule_once(lambda dt: self.callback(status, msg), 0)


class LoadSitesListThread(threading.Thread):

    def __init__(self, client, callback):
        super(LoadSitesListThread, self).__init__()
        self.client = client
        self.callback = callback

    def run(self):
        status, data = self.client.sites()
        # Also retrieving the latest version number
        version_data = None
        if status:
            version_status, version_data = self.client.newest_version()
            if not version_status:
                version_data = None
        Clock.schedule_once(lambda dt: self.callback(status, data, version_data=version_data), 0)


class SyncDirThread(threading.Thread):

    def __init__(self, domain, path, force, client,
                 sync_callback, sync_indicator_callback, stop_sync_callback,
                 network_error_callback, sync_error_callback,
                 protected_file_change_callback):
        super(SyncDirThread, self).__init__()
        self.domain = domain
        self.path = path
        self.force = force
        self.client = client
        self.sync_callback = sync_callback
        self.sync_indicator_callback = sync_indicator_callback
        self.stop_sync_callback = stop_sync_callback
        self.network_error_callback = network_error_callback
        self.sync_error_callback = sync_error_callback
        self.protected_file_change_callback = protected_file_change_callback

    def run(self):
        app = App.get_running_app()
        domain = app.sites_dir_database[self.domain]['domain'].encode('utf-8')
        try:
            status, msg_or_sync_handler = self.client.sync(
                self.network_error_callback, self.sync_error_callback,
                self.protected_file_change_callback, sitename=domain,
                path=self.path, force=self.force,
                sync_indicator_callback=self.sync_indicator_callback,
                stop_sync_callback=self.stop_sync_callback)
        except OSError as e:
            Clock.schedule_once(
                lambda dt: app.show_info_dialog('Filesystem Error', str(e)), 0)
        else:
            Clock.schedule_once(
                lambda dt: self.sync_callback(
                    self.domain, status, msg_or_sync_handler), 0)


###########
# Helpers #
###########

class WebsitesManager(object):

    def __init__(self, sites_dir_database, sites_list_view):
        self._sites_dir_database = sites_dir_database
        self._sites_list_view = sites_list_view
        self._site_views_cache = {}
        self._site_sync_handlers = {}

    def add_or_update_website(self, domain, site_data):
        site_view = None
        if domain in self._site_views_cache:
            site_view = self._site_views_cache[domain]
        else:
            site_view = WebsiteView(domain)

        name = site_data['name'].encode('utf-8')
        site_view.set_name(name)

        stage_status = site_data['stage_status']
        site_view.set_status_color(STATUS_TO_COLOR[stage_status])

        site_dir = None
        if domain in self._sites_dir_database:
            site_dir = self.get_site_dir(domain)
        else:
            self._sites_dir_database[domain] = {}
        self._sites_dir_database[domain].update(site_data)

        site_view.set_site_dir_widgets(site_dir)

        if domain in self._site_sync_handlers:
            site_view.set_sync_btn_text_to_stop()
        else:
            site_view.set_sync_btn_text_to_sync()

        if domain not in self._site_views_cache:
            self._sites_list_view.add_widget(site_view)
            self._site_views_cache[domain] = site_view

    def remove_website(self, domain):
        site_view = self._site_views_cache[domain]
        self._sites_list_view.remove_widget(site_view)
        del self._site_views_cache[domain]

        if domain in self._site_sync_handlers:
            site_sync_thread = self._site_sync_handlers[domain]
            site_sync_thread.stop()
            del self._site_sync_handlers[domain]

    def clear_websites(self):
        self._sites_list_view.clear_widgets()
        self._site_views_cache.clear()

        self.stop_all_threads()

    def stop_all_threads(self):
        for handler in self._site_sync_handlers.values():
            handler.stop()
        self._site_sync_handlers.clear()

    def get_site_name(self, domain):
        return self._sites_dir_database[domain]['name'].encode('utf-8')

    def get_domains(self):
        return self._site_views_cache.keys()

    def get_site_dashboard_url(self, domain):
        return self._sites_dir_database[domain].get(
            'dashboard_url', None)

    def get_site_stage_url(self, domain):
        return self._sites_dir_database[domain].get(
            'stage_url', None)

    ### Site's sync handler ###

    def _delete_site_sync_handler(self, domain):
        handler = self.get_site_sync_handler(domain)
        if handler:
            handler.stop()
            del self._site_sync_handlers[domain]
            site_view = self._site_views_cache[domain]
            site_view.set_sync_btn_text_to_sync()

    def set_site_sync_handler(self, domain, handler):
        # stopping and removing any older syncing handler
        self._delete_site_sync_handler(domain)

        self._site_sync_handlers[domain] = handler
        site_view = self._site_views_cache[domain]
        site_view.set_sync_btn_text_to_stop()

    def get_site_sync_handler(self, domain):
        return self._site_sync_handlers.get(domain, None)

    def stop_site_sync_handler(self, domain):
        self._delete_site_sync_handler(domain)
        self.hide_sync_loading_overlay(domain)

    def show_sync_loading_overlay(self, domain):
        self._site_views_cache[domain].show_sync_loading_overlay()

    def hide_sync_loading_overlay(self, domain):
        self._site_views_cache[domain].hide_sync_loading_overlay()

    ### Site's sync directory ###

    def get_site_dir(self, domain):
        return self._sites_dir_database[domain].get('dir', None)

    def set_site_dir(self, domain, site_dir):
        self._sites_dir_database[domain]['dir'] = site_dir
        self._sites_dir_database.sync()
        site_view = self._site_views_cache[domain]
        site_view.set_site_dir_widgets(site_dir)


################
# Main widgets #
################

class BaseScreen(Screen):
    pass


class LoginWithEmailWidget(BoxLayout):
    email = ObjectProperty(None)
    password = ObjectProperty(None)


class LoginWithTokenWidget(BoxLayout):
    token = ObjectProperty(None)


class LoginScreen(BaseScreen):
    login_type_container = ObjectProperty(None)
    trouble_btn = ObjectProperty(None)
    create_aldryn_id_btn = ObjectProperty(None)
    email = TabTextInput()
    password = TabTextInput()

    aldryn_logo_white_img = resource_path('img/aldryn_logo_white.png')

    def __init__(self, *args, **kwargs):
        super(LoginScreen, self).__init__(*args, **kwargs)
        self.switch_to_email_and_password()

    def switch_to_email_and_password(self):
        container = self.login_type_container
        if container.children:
            container.remove_widget(container.children[0])
        if not hasattr(self, '_login_with_email'):
            self._login_with_email = LoginWithEmailWidget()
        container.add_widget(self._login_with_email)

    def switch_to_token(self):
        container = self.login_type_container
        if container.children:
            container.remove_widget(container.children[0])
        if not hasattr(self, '_login_with_token'):
            self._login_with_token = LoginWithTokenWidget()
        container.add_widget(self._login_with_token)

    def reset_inputs(self):
        if hasattr(self, '_login_with_email'):
            self._login_with_email.email.text = ''
            self._login_with_email.password.text = ''
        if hasattr(self, '_login_with_token'):
            self._login_with_token.token.text = ''


class EmptyScreen(BaseScreen):
    aldryn_logo_img = resource_path('img/aldryn_logo.png')


class SyncScreen(BaseScreen):
    sites_list_view = ObjectProperty(None)

    aldryn_logo_img = resource_path('img/aldryn_logo.png')


##################
# OS integration #
##################

system = platform.system()


def open_in_file_manager(path):
    if system == 'Windows':
        subprocess.Popen(['start', path], shell=True)
    elif system == 'Darwin':
        subprocess.Popen(['open', path])
    else:
        try:
            subprocess.Popen(['xdg-open', path])
        except OSError:
            Logger.exception('Cannot open external file manager')


def notify(title, message, bring_up=False):
    try:
        kwargs = {'title': title, 'message': message}
        if system == "Windows" or system == "Linux":
            kwargs['app_icon'] = os.path.abspath(get_icon_path())
            kwargs['timeout'] = 4
        notification.notify(**kwargs)
        if system == 'Darwin':
            from AppKit import NSApp, NSApplication
            NSApplication.sharedApplication()
            app = NSApp()
            app.requestUserAttention_(0)  # this should bounce the icon
            # the 0 should be replaced by the NSRequestUserAttentionType.NSCriticalRequest
            # imported from NSApplication? TODO
            # https://developer.apple.com/library/mac/documentation/Cocoa/Reference/ApplicationKit/Classes/NSApplication_Class/Reference/Reference.html#//apple_ref/doc/c_ref/NSRequestUserAttentionType
            if bring_up:
                app.activateIgnoringOtherApps_(True)  # this should bring the app to the foreground
    except Exception as e:
        print e
        Logger.exception('Notification error:\n%s' % e)




# -*- coding: utf-8 -*-
from cStringIO import StringIO
import hashlib
import imp
import json
import os
import re
import shutil
import subprocess
import sys
import tarfile
import tempfile
import time
import yaml
import platform


FILENAME_BASIC_RE = re.compile(r'^[a-zA-Z0-9_@]+[a-zA-Z0-9._@-]*\.[a-zA-Z0-9]{1,23}$')
ALLOWED_EXTENSIONS = [
    '.css',
    '.gif',
    '.htc',
    '.htm',
    '.html',
    '.ico',
    '.jpeg',
    '.jpg',
    '.js',
    '.json',
    '.less',
    '.map',
    '.png',
    '.rb',
    '.sass',
    '.scss',
    '.svg',
    '.webm',
    # font formats
    '.eot',
    '.ttf',
    '.woff',
    '.woff2',
    '.otf',
    # text formats
    '.txt',
    '.md',
    '.rst',
    # document formats
    '.pdf',
    '.ps',
    '.djv',
    '.djvu',
    '.doc',
    '.docx',
    '.ppt',
    '.pptx',
    '.xls',
    '.xlsx',
]

BOILERPLATE_REQUIRED = [
    'package-name',
    'identifier',
    'version',
    'templates',
]

BOILERPLATE_REQUIRED_MSG = {
    'package-name': "The specs for boilerplate.json have recently changed. 'package-name' is a new mandatory field.\n"
                    "If you previously already uploaded this Boilerplate without a 'package-name', you can set a "
                    "package-name at https://control.aldryn.com/account/my-boilerplates/.\n",
}

BOILERPLATE_DEPRECATED_FIELDS = [
    'name',
    'description',
    'url',
    'public',
    'license',
    'author',
]

BOILERPLATE_REQUIRED_FILEPATHS = [
    os.path.join('templates', 'base.html'),
]

APP_REQUIRED = [
    'package-name',
    'installed-apps',
]

APP_DEPRECATED_FIELDS = [
    'name',
    'description',
    'url',
    'public',
    'license',
    'author',
    'version',
]

VALID_LICENSE_FILENAMES = [
    'LICENSE.txt',
    'LICENSE',
    'license.txt',
    'license',
]


class ValidationError(Exception):
    pass


def _validate(config, required, path, required_msg=None):
    required_msg = {} if required_msg is None else required_msg
    license_exists = False
    for valid_license_filename in VALID_LICENSE_FILENAMES:
        license_exists |= os.path.exists(
            os.path.join(path, valid_license_filename))
    if not license_exists:
        return (False, "Required LICENSE.txt file not found")
    valid = True
    valid_msg = []
    for thing in required:
        if isinstance(thing, tuple):
            key, values = thing
        else:
            key, values = thing, []

        if key not in config:
            valid = False

            valid_msg.append(required_msg.get(
                key,
                "Required key %r not found in config" % key
            ))

        for subkey in values:
            if subkey not in config[key]:
                valid = (False, "Required sub key %r in %r not found in config" % (subkey, key))
    if valid and not valid_msg:
        valid_msg = ["Configuration file is valid"]
    return valid, '\n'.join(valid_msg)


def _check_deprecated_fields(config, fields):
    return [
        field for field in fields
        if field in config
    ]


def validate_app_config(config, path):
    aldryn_config_path = os.path.abspath(os.path.join(path, 'aldryn_config.py'))
    if os.path.exists(aldryn_config_path):
        tempdir = tempfile.mkdtemp(prefix='tmp_aldryn_client_')
        try:
            shutil.copy(aldryn_config_path, tempdir)
            filepath = os.path.join(tempdir, 'aldryn_config.py')
            try:
                orig_err = sys.stderr
                sys.stderr = StringIO()
                # suppressing "RuntimeWarning: Parent module 'aldryn_config' not found while handling absolute import"
                module = imp.load_source('aldryn_config.config_%s' % int(time.time()), filepath)
                sys.stderr = orig_err
                # checking basic functionality of the Form
                form = module.Form({})
                form.is_valid()
            except Exception:
                import traceback
                error_msg = traceback.format_exc()
                return (False, "Exception in aldryn_config.py:\n\n%s" % error_msg)
        finally:
            shutil.rmtree(tempdir)
    valid, msg = _validate(config, APP_REQUIRED, path)
    # warn about deprecated fields
    depricated_fields = _check_deprecated_fields(config, APP_DEPRECATED_FIELDS)
    if depricated_fields:
        msg += (
            "\n\nDeprecation warning! "
            "These fields are ignored. It's recommended to remove them from addon.json and use the web interface ({0}) to edit them instead.\n"
        ).format('https://control.aldryn.com/account/my-addons/')
        msg += '\n'.join(['  - {0}'.format(field) for field in depricated_fields])
    return valid, msg


def validate_boilerplate_config(config, path):
    for required_filepath in BOILERPLATE_REQUIRED_FILEPATHS:
        dirpath = os.path.join(path, required_filepath)
        if not os.path.exists(dirpath):
            msg = 'Required file "%s" not found' % required_filepath
            return (False, msg)
    (valid, msg) = _validate(config, BOILERPLATE_REQUIRED, path, BOILERPLATE_REQUIRED_MSG)
    if not valid:
        return (False, msg)
    # check templates
    data = config.get('templates', [])
    template_valid = True
    if not isinstance(data, list):
        template_valid = False
    else:
        for template in data:
            if not isinstance(template, list):
                template_valid = False
            elif len(template) != 2:
                template_valid = False
    if not template_valid:
        msg = "Templates must be a list of lists of two items"
        return (False, msg)

    # check protected
    protected = config.get('protected', [])
    valid = True
    if not isinstance(protected, list):
        valid = False
        msg = "Protected files must be a list"
    else:
        errors = []
        for filename in protected:
            filepath = os.path.join(path, filename)
            if filepath.endswith('*'):
                continue
            if not os.path.exists(filepath):
                valid = False
                errors.append("Protected file %r not found" % filename)
        if errors:
            msg = os.linesep.join(errors)
    # warn about deprecated fields
    depricated_fields = _check_deprecated_fields(config, BOILERPLATE_DEPRECATED_FIELDS)
    if depricated_fields:
        msg += (
            "\n\nDeprecation warning! "
            "These fields are ignored. It's recommended to remove them from boilerplate.json and use the web interface ({0}) to edit them instead.\n"
        ).format('https://control.aldryn.com/account/my-boilerplates/')
        msg += '\n'.join(['  - {0}'.format(field) for field in depricated_fields])
    return (valid, msg)


def tar_add_stringio(tar, string_io, name):
    info = tarfile.TarInfo(name=name)
    string_io.seek(0, os.SEEK_END)
    info.size = string_io.tell()
    string_io.seek(0)
    tar.addfile(tarinfo=info, fileobj=string_io)


def is_valid_file_name(name, logger=None):
    if not FILENAME_BASIC_RE.match(name):
        if logger:
            logger("File name %r is not valid, ignoring." % name)
        return False
    ext = os.path.splitext(name)[-1]
    if ext not in ALLOWED_EXTENSIONS:
        if logger:
            logger("File extension %r is not allowed, ignoring." % ext)
        return False
    return True


def filter_static_files(tarinfo):
    if not tarinfo.isfile():
        return tarinfo
    basename = os.path.basename(tarinfo.name)
    if is_valid_file_name(basename):
        return tarinfo
    else:
        return None


def filter_sass_files(tarinfo):
    basename = os.path.basename(tarinfo.name)
    if tarinfo.isfile():
        if is_valid_file_name(basename):
            return tarinfo
        else:
            return None
    elif basename.startswith('.'):
        return None
    else:
        return tarinfo


def filter_template_files(tarinfo):
    if not tarinfo.isfile():
        return tarinfo
    basename = os.path.basename(tarinfo.name)
    ext = os.path.splitext(basename)[1]
    if ext == '.html':
        return tarinfo
    else:
        return None


def _get_license_filename(path):
    for valid_license_filename in VALID_LICENSE_FILENAMES:
        filepath = os.path.join(path, valid_license_filename)
        if os.path.exists(filepath):
            return filepath


def load_license(path):
    filepath = _get_license_filename(path)
    if filepath:
        with open(filepath, 'r') as f:
            license = f.read()
            return license


def bundle_boilerplate(config, path, **complex_extra):
    fileobj = StringIO()
    tar = tarfile.open(mode='w:gz', fileobj=fileobj)
    config_fileobj = StringIO()
    json.dump(config, config_fileobj)
    tar_add_stringio(tar, config_fileobj, 'boilerplate.json')
    data_fileobj = StringIO()
    tar_add_stringio(tar, data_fileobj, 'data.yaml')
    license_filepath = _get_license_filename(path)
    if license_filepath:
        tar.add(license_filepath, 'LICENSE.txt')
    for key, value in complex_extra.items():
        dirpath = os.path.join(path, key)
        if os.path.exists(dirpath):
            tar.add(key, filter=value)
    special_files = [
        f for f in config['protected']
        if not any([f.startswith(key) for key in complex_extra.keys()])
    ]
    for special_file in special_files:
        tar.add(special_file)
    tar.close()
    fileobj.seek(0)
    return fileobj


def get_package_version(path):
    devnull = open(os.devnull, 'w')
    try:
        version = subprocess.check_output(
            ['python', 'setup.py', '--version'], cwd=path, stderr=devnull)
    finally:
        devnull.close()
    return version.strip()


def bundle_package(workspace, tar, path):
    devnull = open(os.devnull, 'w')
    try:
        subprocess.check_call(['python', 'setup.py', 'sdist', '-d', workspace],
                              cwd=path, stdout=devnull, stderr=devnull)
    finally:
        devnull.close()
    egg_file = os.path.join(workspace, os.listdir(workspace)[0])
    tar.add(egg_file, arcname='package.tar.gz')


def bundle_app(config, script, path):
    fileobj = StringIO()
    tar = tarfile.open(mode='w:gz', fileobj=fileobj)
    config_fileobj = StringIO()
    json.dump(config, config_fileobj)
    tar_add_stringio(tar, config_fileobj, 'addon.json')
    script_fileobj = StringIO(script)
    license_filepath = _get_license_filename(path)
    if license_filepath:
        tar.add(license_filepath, 'LICENSE.txt')
    if os.path.exists('aldryn_config.py'):
        tar_add_stringio(tar, script_fileobj, 'aldryn_config.py')
        # add actual package
    distdir = tempfile.mkdtemp(prefix='tmp_aldryn_client_')
    try:
        bundle_package(distdir, tar, path)
    finally:
        shutil.rmtree(distdir)
    version = get_package_version(path)
    version_fobj = StringIO(version)
    info = tarfile.TarInfo(name='VERSION')
    info.size = len(version_fobj.getvalue())
    tar.addfile(info, fileobj=version_fobj)
    tar.close()
    fileobj.seek(0)
    return fileobj


def hashfile(fd, blocksize=65536):
    hasher = hashlib.sha256()
    buf = fd.read(blocksize)
    while len(buf) > 0:
        hasher.update(buf)
        buf = fd.read(blocksize)
    return hasher.digest()


def uniform_filepath(filepath):
    filepath = os.path.abspath(filepath)
    filepath = os.path.realpath(filepath)
    filepath = filepath.rstrip(os.sep)
    return filepath


def is_hidden(path):
    filename = os.path.basename(path)
    return filename.startswith('.')


def is_inside_dir(path, parent_dir):
    path = os.path.join(parent_dir, path)
    path = uniform_filepath(path)
    return path.startswith(parent_dir)


def filter_bad_paths(members, parent_dir):
    parent_dir = uniform_filepath(parent_dir)
    for finfo in members:
        if (not finfo.issym() and not finfo.islnk() and
                is_inside_dir(finfo.path, parent_dir)):
            yield finfo


def resource_path(relative_path):
    """
    Get absolute path of the resource, works for dev and for PyInstaller
    """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = os.sys._MEIPASS
    except Exception:
        base_path = os.path.dirname(__file__)

    return os.path.join(base_path, relative_path)


def get_icon_path():
    system = platform.system()
    if system == 'Darwin':
        icon = 'resources/appIcon.icns'
    elif system == 'Windows':
        icon = 'resources/appIcon.ico'
    else:
        icon = 'resources/appIcon.png'
    return resource_path(icon)


def cli_confirm(question, message=None, default=None):
    if message:
        print message

    if default is None:
        default_answer_str = '[y/n]'
    else:
        default = bool(default)
        if default is True:
            default_answer_str = '[Y/n]'
        elif default is False:
            default_answer_str = '[y/N]'

    question = '%s %s' % (question.rstrip(), default_answer_str)
    while True:
        answer = raw_input(question)
        if answer.lower() == 'y':
            return True
        elif answer.lower() == 'n':
            return False
        elif default is not None and answer == '':
            return default
        else:
            print "Invalid answer, please type either y or n"


def load_boilerplate_config(path):
    from .client import Client
    boilerplate_filename = os.path.join(path, Client.BOILERPLATE_FILENAME_JSON)
    if not os.path.exists(boilerplate_filename):
        msg = "Please provide a %s config file" % Client.BOILERPLATE_FILENAME_JSON
        return (False, msg)
    with open(boilerplate_filename) as fobj:
        try:
            config = json.load(fobj)
        except ValueError as e:
            return (False, repr(e))
        return (True, config,)


def load_app_config(path):
    from .client import Client
    app_filename_json = os.path.join(path, Client.APP_FILENAME_JSON)
    addon_filename_json = os.path.join(path, Client.ADDON_FILENAME_JSON)
    app_filenames = []
    if os.path.exists(app_filename_json):
        app_filenames.append(app_filename_json)
    if os.path.exists(addon_filename_json):
        app_filenames.append(addon_filename_json)
    if len(app_filenames) == 0:
        msg = "File '%s' not found." % Client.ADDON_FILENAME_JSON
        return (False, msg)
    elif len(app_filenames) == 1:
        app_filename = app_filenames[0]
    else:
        msg = "Please provide only one config file (%s)" % (
            ' or '.join(map(os.path.basename, app_filenames)),)
        return (False, msg)
    with open(app_filename) as fobj:
        try:
            config = json.load(fobj)
        except (ValueError) as e:
            return (False, repr(e))
    return (True, config)
